import os
import random
import time

import hiddenlayer as hl
import numpy as np
import pandas as pd
import torch
import torch as t
from sklearn.preprocessing import PolynomialFeatures
from torch import nn, optim
from torch.optim.lr_scheduler import ReduceLROnPlateau

from evaluate.performance import evlauate,evaluationB
from load_data import get_score, import_data
from nModel import DMRModel
from plotfig import GetAUPRC
from RWR import rwr
from randomSeed import GetRandom,threeNumberlist

root_path = 'newseed_result'

torch.cuda.empty_cache()

device = t.device("cuda:0" if torch.cuda.is_available() else 'cpu')
    
class Config(object):
    def __init__(self,v_rwr,v_lr,v_sim,ti):
        self.data_path = '../datasets/DMRLoc_22'
        self.validation = 10
        self.epoch = 300
        self.lr=v_lr
        self.rwr=v_rwr
        self.simtype=v_sim # functional2 # gip
        self.save_path = '../'+root_path+'/'+v_sim+'/'+str(self.validation)+'_fold_rwr_'+ str(v_rwr)+'_sim_'+ v_sim  +'_lr_'+ str(v_lr)  +'_step_'+ str(self.epoch)+'_ti_' + ti +'/'

def set_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

########k fold train valid test ############        
def get_k_fold_node_data(k, i,index):  
    train_mask = torch.ones(len(index), dtype=torch.bool) 
    test_mask = torch.zeros(len(index), dtype=torch.bool) 
    valid_mask = torch.zeros(len(index), dtype=torch.bool) 
    fold_size = len(index) // k  
    for j in range(i*fold_size,(i+1)*fold_size):
        train_mask[index[j]] = False
        valid_mask[index[j]] = True
    if i == 9:
        for j in range(0,fold_size):
            train_mask[index[j]] = False
            test_mask[index[j]] = True
    else:
        for j in range((i+1)*fold_size,(i+2)*fold_size):
            train_mask[index[j]] = False
            test_mask[index[j]] = True
    return  train_mask,valid_mask,test_mask

def device_info():
    # prettytable 
    import platform
    import sys

    import prettytable as pt
    import torch
    tb = pt.PrettyTable()
    tb.field_names = ["Items", "INFO"]
    tb.add_row(["python version",platform.python_version()])
    tb.add_row(["pytorch version",torch.__version__ ])
    tb.add_row(["cuda version",torch.version.cuda])
    tb.add_row(["cudnn version",torch.backends.cudnn.version()])
    tb.add_row(["cuda available",torch.cuda.is_available()])
    tb.add_row(["device count",torch.cuda.device_count()])
    print(tb)

def train(model, train_data, opt,history):
    print('Running train model ...')
    eval_loss_list = []
    eval_acc_list = []
    criterion =  nn.BCELoss()  # MultiLabelSoftMarginLoss CrossEntropyLoss
    loss_func = nn.MSELoss()
    # criterion = nn.BCEWithLogitsLoss(reduction='none')
    optimizer = optim.SGD(model.parameters(), lr=opt.lr,momentum=0.9) # optim.SGD(mnist_net.parameters(), lr=0.002)
    scheduler = ReduceLROnPlateau(optimizer, 'min')
    def train_epoch(epoch):
        model.train()
        target,label,encoded, decoded  = model(train_data["pi_feature"],
                        train_data["dis_feature"],
                        train_data["pi_sim"],
                        train_data["dis_sim"],
                        train_data["relation"],
                        train_data["miRNAseqSim"]
                        )
        train_loss= criterion(target[train_data['train_label']], train_data['ture_label'][train_data['train_label']].to(device)  ) + loss_func(encoded, decoded)
        # train_loss= loss_func(encoded, decoded)
        # train_loss = (train_loss * weights).mean()
        # val_loss2 = loss_func(encoded, decoded)
        validation_loss = validation(model,train_data,opt)
        optimizer.zero_grad()
        train_loss.backward()
        # val_loss2.backward()
        optimizer.step()
        scheduler.step(validation_loss)
        # 在训练集上预测并计算正确率
        pred_label = np.array(label[train_data['train_label']].cpu().detach().numpy(),dtype=int)
        pred_score = np.array(target[train_data['train_label']].cpu().detach().numpy(),dtype=float)
        ture_label = np.array(train_data['ture_label'][train_data['train_label']].cpu().detach().numpy(),dtype=int)
        myMetic = evlauate(pred_label,ture_label,pred_score)
        print("train loss:{0}, validation_loss loss:{1}".format(train_loss,validation_loss))
        history.log((epoch),
                    t_loss=train_loss,
                    v_loss = validation_loss,
                    average_precision = myMetic[0],
                    coverage_error = myMetic[1],
                    ranking_loss = myMetic[2],
                    hamming_loss = myMetic[3],
                    zero_one_loss = myMetic[4],
                    acc_score = myMetic[5]
                    )    
        return train_loss.item()
        
    for epoch in range(1, opt.epoch+1):
        train_reg_loss = train_epoch(epoch)
        eval_loss_list.append(train_reg_loss)

    apMetic,apMetic_v1,auc_score, aupr_score = test(model,train_data,opt)

    return  apMetic,apMetic_v1,auc_score, aupr_score

def validation(model,train_data,opt):
    model.eval()
    fold_num =train_data['fold_num']
    criterion =  nn.BCELoss() 
    loss_func = nn.MSELoss()
    # criterion = nn.BCEWithLogitsLoss(reduction='none')
    with torch.no_grad():
        target,label,encoded, decoded = model(train_data["pi_feature"],
                        train_data["dis_feature"],
                        train_data["pi_sim"],
                        train_data["dis_sim"],
                        train_data["relation"],
                        train_data["miRNAseqSim"])
        val_loss1 = criterion(target[train_data['valid_label']], train_data['ture_label'][train_data['valid_label']].to(device))
        val_loss2 = loss_func(encoded, decoded)
        # val_loss = (val_loss * weights).mean()
        pred_label = np.array(label[train_data['valid_label']].cpu(),dtype=int)
        pred_score = np.array(target[train_data['valid_label']].cpu(),dtype=float)
        ture_label = np.array(train_data['ture_label'][train_data['valid_label']].cpu(),dtype=int)
        
        np.savetxt(opt.save_path+str(fold_num)+'_valid_pred_label.txt', pred_label, fmt="%d", delimiter=",")
        np.savetxt(opt.save_path+str(fold_num)+'_valid_pred_score.txt', pred_score, fmt="%0.4f", delimiter=",")
        np.savetxt(opt.save_path+str(fold_num)+'_valid_ture_label.txt', ture_label, fmt="%d", delimiter=",")
           
    return  val_loss2 + val_loss1

def test(model,train_data,opt):
    print('Running test model ...')
    model.eval()
    fold_num =train_data['fold_num']
    # criterion = nn.MultiLabelMarginLoss() # nn.MultiLabelMarginLoss()
    # weights = torch.tensor([0.16, 0.16, 0.25, 0.25, 0.083, 0.083,0.031]).to(device)
    # criterion = nn.BCEWithLogitsLoss(reduction='none')
    criterion =  nn.BCELoss() 
    loss_func = nn.MSELoss()
    with torch.no_grad():
        target,label,encoded, decoded  = model(train_data["pi_feature"],
                        train_data["dis_feature"],
                        train_data["pi_sim"],
                        train_data["dis_sim"],
                        train_data["relation"],
                        train_data["miRNAseqSim"])
        # test_loss = criterion(target[train_data['test_label']], train_data['ture_label'][train_data['test_label']].to(device))
        pred_label = np.array(label[train_data['test_label']].cpu(),dtype=int)
        pred_score = np.array(target[train_data['test_label']].cpu(),dtype=float)
        ture_label = np.array(train_data['ture_label'][train_data['test_label']].cpu(),dtype=int)
        
        auc_score, aupr_score = get_score(ture_label,pred_score)
        print(auc_score, aupr_score)
        feature = model.feature
        feature = np.array(feature.cpu(),dtype=float)
        np.savetxt(opt.save_path+str(fold_num)+'_test_pred_feature.txt', feature, fmt="%0.10f", delimiter=",")
        np.savetxt(opt.save_path+str(fold_num)+'_test_pred_label.txt', pred_label, fmt="%d", delimiter=",")
        np.savetxt(opt.save_path+str(fold_num)+'_test_pred_score.txt', pred_score, fmt="%0.4f", delimiter=",")
        np.savetxt(opt.save_path+str(fold_num)+'_test_ture_label.txt', ture_label, fmt="%d", delimiter=",")
        myMetic = evlauate(pred_label,ture_label,pred_score)
        myMeticv1 = evaluationB(ture_label,pred_label)
        print(myMeticv1)
    return myMetic,myMeticv1,auc_score, aupr_score

def main(opt,seed):
    set_seed(seed)
    os.makedirs(opt.save_path)
    didi_matrix,mimi_matrix,dimi_matrix,ben_ind_label,smith_waterman_gip_miRNAsim,n_didi,n_mimi = import_data(opt.data_path,opt.simtype)
    index = [i for i  in range(n_mimi)]
    random.shuffle(index)
    _pi_features = rwr(mimi_matrix, opt.rwr)
    poly_dis = PolynomialFeatures(1, include_bias=False)
    pi_features = poly_dis.fit_transform(_pi_features)
    print(pi_features.shape)
    # extract disease features by Random Walk with Restart and raise dimension by PolynomialFeatures
    _dis_features = rwr(didi_matrix, opt.rwr)
    poly_dis = PolynomialFeatures(2, include_bias=False)
    dis_features = poly_dis.fit_transform(_dis_features)
    print(dis_features.shape)
    pi=np.zeros((n_mimi, n_mimi))
    row_id, col_id = np.diag_indices_from(pi)
    pi[row_id, col_id] = 1

    dis=np.zeros((n_didi, n_didi))
    row_id, col_id = np.diag_indices_from(dis)
    dis[row_id, col_id] = 1
    rel = np.concatenate((
            np.concatenate((pi, dimi_matrix), axis=1),
            np.concatenate((dimi_matrix.T, dis), axis=1)
        ), axis=0)
    filename = str(opt.validation) +'-fold-score'
    index = [i for i  in range(n_mimi)]
    random.shuffle(index)
    average_precision_list = []
    coverage_error_list = []
    ranking_loss_list = []
    hamming_loss_list = []
    zero_one_loss_list = []
    acc_score_list = []
    n_estimators_list = []
    auc_score_list = []
    aupr_score_list = []
    ######################
    sum_precision = []
    sum_recall = []
    sum_accuracy = []
    sum_f1_scor = []
    sum_micro_fscore = []
    sum_macro_fscore = []
    sum_hamming_loss = []
    sum_averagePrecision = []

    filename = str(opt.validation) +'-fold-mutilabelscore'
    threshold = t.tensor([0.50]).to(device)

    # model = nn.DataParallel(model).to(device)
    for i in range(opt.validation):
        print('Running {} fold ...'.format(i+1))
        train_mask,valid_mask,test_mask = get_k_fold_node_data(opt.validation, i,index)
        train_data = {
        "pi_feature": t.FloatTensor(pi_features).to(device),
        "dis_feature": t.FloatTensor(dis_features).to(device),
        "pi_sim": t.FloatTensor(mimi_matrix).to(device),
        "dis_sim": t.FloatTensor(didi_matrix).to(device),
        "train_label": train_mask.to(device),
        "relation": t.FloatTensor(rel).to(device),
        "test_label": test_mask.to(device),
        "valid_label": valid_mask.to(device),
        "ture_label": t.FloatTensor(ben_ind_label).to(device),
        "fold_num":i,
        'miRNAseqSim':t.FloatTensor(smith_waterman_gip_miRNAsim).to(device)
        }
        history = hl.History()
        model = DMRModel(n_mimi,n_didi,pi_features,dis_features,threshold).to(device)
        apMetic,evaluation_measures,auc_score, aupr_score = train(model, train_data, opt,history)

        n_estimators_list.append(i+1)
        average_precision_list.append(apMetic[0])
        coverage_error_list.append(apMetic[1])
        ranking_loss_list.append(apMetic[2])
        hamming_loss_list.append(apMetic[3])
        zero_one_loss_list.append(apMetic[4])
        acc_score_list.append(apMetic[5])
        auc_score_list.append(list(auc_score))
        aupr_score_list.append(list(aupr_score))

        sum_precision.append(evaluation_measures["Precision"])
        sum_recall.append(evaluation_measures["Recall"])
        sum_accuracy.append(evaluation_measures["Accuracy"])
        sum_f1_scor.append(evaluation_measures["F1 score"])
        sum_micro_fscore.append(evaluation_measures["f-1 Micro"])
        sum_macro_fscore.append(evaluation_measures["f-1 Macro"])
        sum_hamming_loss.append(evaluation_measures["Hamming Loss"])
        sum_averagePrecision.append(evaluation_measures["averagePrecision"])

        np.savetxt(opt.save_path+str(i)+'_test_auc_score.txt', auc_score, fmt="%f", delimiter=",")
        np.savetxt(opt.save_path+str(i)+'_test_aupr_score.txt', aupr_score, fmt="%f", delimiter=",")   
        history.save(opt.save_path+"{}-history.pkl".format(i))
        fold = 10
    with open(opt.save_path+'/cnn_results', 'a', newline='') as file:
        import csv
        writer = csv.writer(file)
        for index in range(fold):
            writer.writerow(
            [str(index), str(round(sum_precision[index], 4)), str(round(sum_recall[index] ,4)), str(round(sum_f1_scor[index] ,4)),
            str(round(sum_micro_fscore[index] ,4)), str(round(sum_macro_fscore[index] ,4)), str(round(sum_accuracy[index] ,4)),
            str(round(sum_hamming_loss[index] ,4)), str(round(sum_averagePrecision[index],4))])
        writer.writerow(
            ["rand", str(round(sum(sum_precision) / fold, 4)), str(round(sum(sum_recall) / fold,4)), str(round(sum(sum_f1_scor) / fold,4)),
            str(round(sum(sum_micro_fscore) / fold,4)), str(round(sum(sum_macro_fscore) / fold,4)), str(round(sum(sum_accuracy) / fold,4)),
            str(round(sum(sum_hamming_loss) / fold,4)), str(round(sum(sum_averagePrecision) / fold,4))])

    # save score
    df = pd.DataFrame(
                {'k_fold_number': n_estimators_list,
                 'Average-Precision': average_precision_list,
                 'Accurcy': acc_score_list,
                 'Coverage': coverage_error_list,
                 'Ranking-loss': ranking_loss_list,
                 'Hamming-loss': hamming_loss_list,
                 'One-error': zero_one_loss_list})
    df.to_csv(opt.save_path + filename + ".csv",
                index=False, float_format='%.5f')

    df = pd.DataFrame(
            {'k_fold_number': n_estimators_list,
                'auc_score': auc_score_list,
               })
    df.to_csv(opt.save_path+ str(opt.validation) + '-fold-auc_score' + ".csv",
                index=False, float_format='%.5f')
              
    df = pd.DataFrame(
        {'k_fold_number': n_estimators_list,
            'aupr_score': aupr_score_list,
            })
    df.to_csv(opt.save_path+ str(opt.validation)  +'-fold-aupr_score' + ".csv",
                index=False, float_format='%.5f')
    # torch.cuda.empty_cache()


if __name__ == "__main__":
    '''
    0.7 
    '''
    rwr_list = [0.7]
    lr_list = [0.0004]
    seed = GetRandom() #
    simtype_list = ["functional1"]
    device_info()
    root_path = "All_109_Seed_v4"
    for simtype in simtype_list:
        roc_list = []
        aucpr_list = []
        for sd in seed:
            for v_rwr in rwr_list:
                for v_lr in lr_list:
                    ti = str(sd) +'_'+ str(time.strftime("%H-%M-%S")) #  str(time.strftime("%H-%M-%S"))  #  str(sd) #  
                    opt = Config(v_rwr,v_lr,simtype,ti)
                    print(opt.lr,opt.rwr,opt.simtype)
                    roc,aucpr =  main(opt,sd)
                    roc =  list(roc)
                    aucpr = list(aucpr) 
                    roc.insert(0,v_rwr)
                    roc.insert(0,v_lr)
                    roc.insert(0,sd)
                    aucpr.insert(0,v_rwr)
                    aucpr.insert(0,v_lr)
                    aucpr.insert(0,sd)
                    roc_list.append(roc)
                    aucpr_list.append(aucpr)
                # torch.cuda.set_device(1)
                # torch.cuda.empty_cache()kill
        roc = np.array(roc_list)
        np.savetxt("../"+root_path+"/"+ simtype + '/localization_AUC.txt', roc, fmt="%f", delimiter=",")
        aucpr = np.array(aucpr_list)
        np.savetxt("../"+root_path+"/"+ simtype + '/localization_PR.txt', aucpr, fmt="%f", delimiter=",")
