import torch as t
from torch import nn
import torch.nn.functional as F
from torch_geometric.nn import conv,dense, norm
from torch import nn, optim
# t.manual_seed(66)


class autoencoder(nn.Module):
    def __init__(self, input_dim):
        super(autoencoder, self).__init__()
        self.encoder = nn.Sequential(
            nn.Linear(input_dim, 32),
            nn.ReLU()
        )

        self.decoder= nn.Sequential(
            nn.Linear(32, input_dim),
            nn.Sigmoid())

    def forward(self, input):
        encoded = self.encoder(input)
        score = self.decoder(encoded)
        return input,score


class DMRModel(nn.Module):
    def __init__(self, n_pi, n_dis,X_pi, X_dis,threshold):
        super(DMRModel, self).__init__()

        self.bm_input_x = norm.BatchNorm(X_pi.shape[1])
        self.bm_input_y = norm.BatchNorm(X_dis.shape[1])

        self.feature = None

        self.dense_x = nn.Linear(X_pi.shape[1], 800)
        self.dense_y = nn.Linear(X_dis.shape[1], 800)

        self.bm_x0 = norm.BatchNorm(800)
        self.bm_y0 = norm.BatchNorm(800)
        self.autoencoder = autoencoder(X_pi.shape[1])

        #Asso-GCN layer1
        self.gcn_xy1 = dense.DenseGraphConv(800, 800)

        self.gcn_x1 = dense.DenseGCNConv(800, 800)

        self.gcn_y1 = dense.DenseGCNConv(800, 800)
        
        self.bm_xy1 = norm.BatchNorm(800)
        self.bm_x1 = norm.BatchNorm(800)
        self.bm_y1 = norm.BatchNorm(800)
        
        #Asso-GCN layer2
        # self.gcn_xy2 = dense.DenseGraphConv(800, 800)
        # self.gcn_x2 = dense.DenseGCNConv(800, 800)
        # self.gcn_y2 = dense.DenseGCNConv(800, 800)
        # self.bm_xy2 = norm.BatchNorm(800)
        # self.bm_x2 = norm.BatchNorm(800)
        # self.bm_y2 = norm.BatchNorm(800)

        #Asso-GCN layer3
        # self.gcn_xy3 = dense.DenseGraphConv(800, 800)
        # self.gcn_x3 = dense.DenseGCNConv(800, 800)
        # self.gcn_y3 = dense.DenseGCNConv(800, 800)
        # self.bm_xy3 = norm.BatchNorm(800)
        # self.bm_x3 = norm.BatchNorm(800)
        # self.bm_y3 = norm.BatchNorm(800)

        #Asso-GCN layer4
        # self.gcn_xy4 = dense.DenseGraphConv(800, 800)
        # self.gcn_x4 = dense.DenseGCNConv(800, 800)
        # self.gcn_y4 = dense.DenseGCNConv(800, 800)
        # self.bm_xy4 = norm.BatchNorm(800)
        # self.bm_x4 = norm.BatchNorm(800)
        # self.bm_y4 = norm.BatchNorm(800)

        #Asso-GCN layer5
        # self.gcn_xy5 = dense.DenseGraphConv(800, 800)
        # self.gcn_x5 = dense.DenseGCNConv(800, 800)
        # self.gcn_y5 = dense.DenseGCNConv(800, 800)
        # self.bm_xy5 = norm.BatchNorm(800)
        # self.bm_x5 = norm.BatchNorm(800)
        # self.bm_y5 = norm.BatchNorm(800)

        self.linear_x1 = nn.Linear(800, 400)
        self.linear_y1 = nn.Linear(800, 400)
        self.linear_x2 = nn.Linear(400, 200)
        self.linear_y2 = nn.Linear(400, 200)
        self.linear_x3 = nn.Linear(200, 100)
        self.linear_y3 = nn.Linear(200, 100)
       
        self.linear_x4 = nn.Linear(100, 50)
        self.linear_y4 = nn.Linear(100, 50)

        self.bm_lx1 = norm.BatchNorm(400)
        self.bm_ly1 = norm.BatchNorm(400)
        self.bm_lx2 = norm.BatchNorm(200)
        self.bm_ly2 = norm.BatchNorm(200)
        self.bm_lx3 = norm.BatchNorm(100)
        self.bm_ly3 = norm.BatchNorm(100)
      
        self.dense_out = nn.Linear(2836, 800)
        self.pre_out = nn.Linear(n_pi+n_dis, 7)
        self.bn_in = norm.BatchNorm(n_dis)
        self.bn_out = norm.BatchNorm(7)
        self.MaxPool = nn.MaxPool1d(3, stride=1)

        self.threshold = threshold # t.tensor([0.5]).cuda()


    def forward(self,  X_pi, X_dis, A_pi, A_dis, A_rel,SeqSim):

        m, n = X_pi.shape[0], X_dis.shape[0]

        X = self.bm_input_x(X_pi)
        Y = self.bm_input_y(X_dis)
        X = F.leaky_relu(self.dense_x(X))
        Y = F.leaky_relu(self.dense_y(Y))

        X = self.bm_x0(X)
        Y = self.bm_y0(Y)

        #Asso-GCN layer1
        XY = t.squeeze(F.leaky_relu(self.gcn_xy1(t.cat((X, Y)), A_rel)))
        XY = self.bm_xy1(XY)
        X, Y = t.split(XY, (m, n))
        #miRNA-GCN layer1
        X = t.squeeze(F.leaky_relu(self.gcn_x1(X, A_pi)))
        #disease-GCN layer1
        Y = t.squeeze(F.leaky_relu(self.gcn_y1(Y, A_dis)))
        X = self.bm_x1(X)
        Y = self.bm_y1(Y)

        # #Asso-GCN layer2
        # XY = t.squeeze(F.leaky_relu(self.gcn_xy2(t.cat((X, Y)), A_rel)))
        # XY = self.bm_xy2(XY)
        # X, Y = t.split(XY, (m, n))

        # #Sim-GCN layer2
        # X = t.squeeze(F.leaky_relu(self.gcn_x2(X, A_pi)))
        # Y = t.squeeze(F.leaky_relu(self.gcn_y2(Y, A_dis)))
        # X = self.bm_x2(X)
        # Y = self.bm_y2(Y)

        # #Asso-GCN layer3
        # XY = t.squeeze(F.leaky_relu(self.gcn_xy3(t.cat((X, Y)), A_rel)))
        # XY = self.bm_xy3(XY)
        # X, Y = t.split(XY, (m, n))

        # #Sim-GCN layer3
        # X = t.squeeze(F.leaky_relu(self.gcn_x3(X, A_pi)))
        # Y = t.squeeze(F.leaky_relu(self.gcn_y3(Y, A_dis)))
        # X = self.bm_x3(X)
        # Y = self.bm_y3(Y)

        # #Asso-GCN layer4
        # XY = t.squeeze(F.leaky_relu(self.gcn_xy4(t.cat((X, Y)), A_rel)))
        # XY = self.bm_xy4(XY)
        # X, Y = t.split(XY, (m, n))

        # #Sim-GCN layer4
        # X = t.squeeze(F.leaky_relu(self.gcn_x4(X, A_pi)))
        # Y = t.squeeze(F.leaky_relu(self.gcn_y4(Y, A_dis)))
        # X = self.bm_x4(X)
        # Y = self.bm_y4(Y)

        # #Asso-GCN layer5
        # XY = t.squeeze(F.leaky_relu(self.gcn_xy5(t.cat((X, Y)), A_rel)))
        # XY = self.bm_xy5(XY)
        # X, Y = t.split(XY, (m, n))

        # #Sim-GCN layer5
        # X = t.squeeze(F.leaky_relu(self.gcn_x5(X, A_pi)))
        # Y = t.squeeze(F.leaky_relu(self.gcn_y5(Y, A_dis)))
        # X = self.bm_x5(X)
        # Y = self.bm_y5(Y)

        X = F.relu(self.linear_x1(X))
        Y = F.relu(self.linear_y1(Y))
        X = self.bm_lx1(X)
        Y = self.bm_ly1(Y)
        X = F.relu(self.linear_x2(X))
        Y = F.relu(self.linear_y2(Y))
        X = self.bm_lx2(X)
        Y = self.bm_ly2(Y)
        X = F.relu(self.linear_x3(X))
        Y = F.relu(self.linear_y3(Y))

        # X = self.bm_lx3(X)
        # Y = self.bm_ly3(Y)
        # X = F.relu(self.linear_x4(X))
        # Y = F.relu(self.linear_y4(Y))
        
        mirnadiseasefeature = X.mm(Y.t())
        sequencesencoded,sequencesdecoded = self.autoencoder(SeqSim)
        # label = t.squeeze(score)
        
        # print(sequencesfeature.shape,mirnadiseasefeature.shape)
        feature = t.cat((mirnadiseasefeature,sequencesdecoded),1)
        self.feature = feature
        # out =  t.squeeze(F.leaky_relu(self.dense_out(score)))
        label = self.pre_out(feature)
        pre_score = t.sigmoid(label)

        pre_label = (pre_score>self.threshold).float()*1

        return pre_score, pre_label,sequencesencoded,sequencesdecoded



# #Asso-GCN layer2
# XY = t.squeeze(F.leaky_relu(self.gcn_xy2(t.cat((X, Y)), A_rel)))
# XY = self.bm_xy2(XY)
# X, Y = t.split(XY, (m, n))

# #Sim-GCN layer2
# X = t.squeeze(F.leaky_relu(self.gcn_x2(X, A_pi)))
# Y = t.squeeze(F.leaky_relu(self.gcn_y2(Y, A_dis)))
# X = self.bm_x2(X)
# Y = self.bm_y2(Y)

# #Asso-GCN layer3
# XY = t.squeeze(F.leaky_relu(self.gcn_xy3(t.cat((X, Y)), A_rel)))
# XY = self.bm_xy3(XY)
# X, Y = t.split(XY, (m, n))

# #Sim-GCN layer3
# X = t.squeeze(F.leaky_relu(self.gcn_x3(X, A_pi)))
# Y = t.squeeze(F.leaky_relu(self.gcn_y3(Y, A_dis)))
# X = self.bm_x3(X)
# Y = self.bm_y3(Y)

# #Asso-GCN layer4
# XY = t.squeeze(F.leaky_relu(self.gcn_xy4(t.cat((X, Y)), A_rel)))
# XY = self.bm_xy4(XY)
# X, Y = t.split(XY, (m, n))

# #Sim-GCN layer4
# X = t.squeeze(F.leaky_relu(self.gcn_x4(X, A_pi)))
# Y = t.squeeze(F.leaky_relu(self.gcn_y4(Y, A_dis)))
# X = self.bm_x4(X)
# Y = self.bm_y4(Y)

# #Asso-GCN layer5
# XY = t.squeeze(F.leaky_relu(self.gcn_xy5(t.cat((X, Y)), A_rel)))
# XY = self.bm_xy5(XY)
# X, Y = t.split(XY, (m, n))

# #Sim-GCN layer5
# X = t.squeeze(F.leaky_relu(self.gcn_x5(X, A_pi)))
# Y = t.squeeze(F.leaky_relu(self.gcn_y5(Y, A_dis)))
# X = self.bm_x5(X)
# Y = self.bm_y5(Y)
